from os import PathLike
from unittest.mock import patch

import pytest

from gso.products.product_types.iptrunk import Iptrunk
from gso.products.product_types.router import Router
from gso.utils.shared_enums import Vendor
from test import USER_CONFIRM_EMPTY_FORM
from test.conftest import UseJuniperSide
from test.services.conftest import MockedSharePointClient
from test.workflows import (
    assert_complete,
    assert_lso_interaction_success,
    assert_suspended,
    extract_state,
    resume_workflow,
    run_workflow,
)
from test.workflows.iptrunk.test_create_iptrunk import MockedNetboxClient


@pytest.fixture()
def migrate_form_input(
    request,
    faker,
    iptrunk_subscription_factory,
    router_subscription_factory,
    iptrunk_side_subscription_factory,
):
    use_juniper = getattr(request, "param", UseJuniperSide.NONE)
    new_side_ae_members_nokia = faker.link_members_nokia()[0:2]
    new_side_ae_members_juniper = faker.link_members_juniper()[0:2]

    if use_juniper == UseJuniperSide.SIDE_A:
        # Nokia -> Juniper
        product_id = iptrunk_subscription_factory()
        old_subscription = Iptrunk.from_subscription(product_id)
        new_router = router_subscription_factory(vendor=Vendor.JUNIPER)
        replace_side = str(old_subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.subscription.subscription_id)
        new_side_ae_members = new_side_ae_members_juniper
        lag_name = "ae1"
    elif use_juniper == UseJuniperSide.SIDE_B:
        # Juniper -> Nokia
        old_side_a_node = router_subscription_factory(vendor=Vendor.JUNIPER)
        old_side_a_node = iptrunk_side_subscription_factory(iptrunk_side_node=old_side_a_node)
        old_side_b_node = router_subscription_factory(vendor=Vendor.JUNIPER)
        old_side_b_node = iptrunk_side_subscription_factory(iptrunk_side_node=old_side_b_node)
        product_id = iptrunk_subscription_factory(iptrunk_sides=[old_side_a_node, old_side_b_node])
        old_subscription = Iptrunk.from_subscription(product_id)
        new_router = router_subscription_factory()
        replace_side = str(old_subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.subscription.subscription_id)
        new_side_ae_members = new_side_ae_members_nokia
        lag_name = "lag-1"
    elif use_juniper == UseJuniperSide.SIDE_BOTH:
        # Juniper -> Juniper
        old_side_a_node = router_subscription_factory(vendor=Vendor.JUNIPER)
        old_side_a_node = iptrunk_side_subscription_factory(iptrunk_side_node=old_side_a_node)
        old_side_b_node = router_subscription_factory(vendor=Vendor.JUNIPER)
        old_side_b_node = iptrunk_side_subscription_factory(iptrunk_side_node=old_side_b_node)
        product_id = iptrunk_subscription_factory(iptrunk_sides=[old_side_a_node, old_side_b_node])
        old_subscription = Iptrunk.from_subscription(product_id)
        new_router = router_subscription_factory(vendor=Vendor.JUNIPER)
        replace_side = str(old_subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.subscription.subscription_id)
        new_side_ae_members = new_side_ae_members_juniper
        lag_name = "ae1"
    else:
        # Nokia -> Nokia
        product_id = iptrunk_subscription_factory()
        old_subscription = Iptrunk.from_subscription(product_id)
        new_router = router_subscription_factory()
        replace_side = str(old_subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.subscription.subscription_id)
        new_side_ae_members = new_side_ae_members_nokia
        lag_name = "lag-1"

    return [
        {"subscription_id": product_id},
        {
            "tt_number": faker.tt_number(),
            "replace_side": replace_side,
            "migrate_to_different_site": True,
        },
        {
            "new_node": new_router,
        },
        {
            "new_lag_interface": lag_name,
            "new_lag_member_interfaces": new_side_ae_members,
        },
    ]


def interface_lists_are_equal(list1, list2):
    if len(list1) != len(list2):
        return False

    for item1 in list1:
        if not any(
            item1.interface_name == item2.interface_name and item1.interface_description == item2.interface_description
            for item2 in list2
        ):
            return False

    return True


@pytest.mark.parametrize(
    "migrate_form_input",
    [UseJuniperSide.NONE, UseJuniperSide.SIDE_A, UseJuniperSide.SIDE_B, UseJuniperSide.SIDE_BOTH],
    indirect=True,
)
@pytest.mark.workflow()
@patch("gso.services.infoblox.create_host_by_ip")
@patch("gso.services.infoblox.delete_host_by_ip")
@patch("gso.services.lso_client._send_request")
@patch("gso.services.netbox_client.NetboxClient.get_available_interfaces")
@patch("gso.services.netbox_client.NetboxClient.get_available_lags")
@patch("gso.services.netbox_client.NetboxClient.create_interface")
@patch("gso.services.netbox_client.NetboxClient.attach_interface_to_lag")
@patch("gso.services.netbox_client.NetboxClient.reserve_interface")
@patch("gso.services.netbox_client.NetboxClient.allocate_interface")
@patch("gso.services.netbox_client.NetboxClient.free_interface")
@patch("gso.services.netbox_client.NetboxClient.delete_interface")
@patch("gso.workflows.iptrunk.migrate_iptrunk.SharePointClient")
def test_migrate_iptrunk_success(
    mock_sharepoint_client,
    mocked_delete_interface,
    mocked_free_interface,
    mocked_allocate_interface,
    mocked_reserve_interface,
    mocked_attach_interface_to_lag,
    mocked_create_interface,
    mocked_get_available_lags,
    mocked_get_available_interfaces,
    mock_execute_playbook,
    mock_delete_host_by_ip,
    mock_create_host_by_ip,
    migrate_form_input,
    data_config_filename: PathLike,
):
    #  Set up mock return values
    mocked_netbox = MockedNetboxClient()
    mocked_get_available_interfaces.return_value = mocked_netbox.get_available_interfaces()
    mocked_attach_interface_to_lag.return_value = mocked_netbox.attach_interface_to_lag()
    mocked_reserve_interface.return_value = mocked_netbox.reserve_interface()
    mocked_allocate_interface.return_value = mocked_netbox.allocate_interface()
    mocked_free_interface.return_value = mocked_netbox.free_interface()
    mocked_create_interface.return_value = mocked_netbox.create_interface()
    mocked_get_available_lags.return_value = mocked_netbox.get_available_lags()
    mocked_delete_interface.return_value = mocked_netbox.delete_interface()
    mock_sharepoint_client.return_value = MockedSharePointClient

    result, process_stat, step_log = run_workflow("migrate_iptrunk", migrate_form_input)

    for _ in range(8):
        result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    assert_suspended(result)
    result, step_log = resume_workflow(process_stat, step_log, input_data=USER_CONFIRM_EMPTY_FORM)

    for _ in range(8):
        result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    assert_suspended(result)
    result, step_log = resume_workflow(process_stat, step_log, input_data=USER_CONFIRM_EMPTY_FORM)

    for _ in range(1):
        result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    #  Continue workflow after it has displayed a checklist URL.
    assert_suspended(result)
    result, step_log = resume_workflow(process_stat, step_log, input_data=USER_CONFIRM_EMPTY_FORM)

    assert_complete(result)

    state = extract_state(result)
    subscription_id = state["subscription_id"]
    subscription = Iptrunk.from_subscription(subscription_id)

    assert subscription.status == "active"
    assert mock_execute_playbook.call_count == 17
    assert mock_create_host_by_ip.call_count == 1
    assert mock_delete_host_by_ip.call_count == 1

    # get some values from form
    new_router = migrate_form_input[2]["new_node"]
    new_lag_interface = migrate_form_input[3]["new_lag_interface"]
    replace_side = migrate_form_input[1]["replace_side"]
    new_lag_member_interfaces = migrate_form_input[3]["new_lag_member_interfaces"]

    # Get vendor for the new and old migrated node
    vendor_old = Router.from_subscription(replace_side).router.vendor
    vendor_new = Router.from_subscription(new_router).router.vendor

    # Only Nokia will be checked on netbox
    num_nokia_lags = 1 if vendor_new == Vendor.NOKIA else 0
    num_nokia_reserved = 2 * (vendor_new == Vendor.NOKIA)
    num_nokia_attached = 2 * (vendor_new == Vendor.NOKIA)

    # Only interfaces lag delete for nokia node is tested
    num_nokia_lag_del = 1 * (vendor_old == Vendor.NOKIA)

    # Only free interfaces when node was nokia
    num_nokia_free = 2 * (vendor_old == Vendor.NOKIA)

    # Assert all Netbox calls have been made
    assert mocked_create_interface.call_count == num_nokia_lags  # once for creating the LAG on the newly replaced side:
    assert mocked_reserve_interface.call_count == num_nokia_reserved  # Twice for the new interfaces
    assert mocked_attach_interface_to_lag.call_count == num_nokia_attached  # Twice for the new interfaces
    assert mocked_allocate_interface.call_count == num_nokia_attached  # Twice for the new interfaces
    assert mocked_free_interface.call_count == num_nokia_free  # Twice for the old interfaces
    assert mocked_delete_interface.call_count == num_nokia_lag_del  # once for deleting the LAG on the old replaced side

    # Assert the new side is replaced
    assert str(subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.subscription.subscription_id) == new_router
    assert subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_iface == new_lag_interface
    existing_members = subscription.iptrunk.iptrunk_sides[0].iptrunk_side_ae_members
    assert interface_lists_are_equal(new_lag_member_interfaces, existing_members)
